require 'common.util'
require 'common.fbo'
require 'common.tex'
require 'common.shader'

function defines(...)
	local lines = {...}
	local str = ''
	for i,line in pairs(lines) do
		str = str .. '#define ' ..line .. '\n'
	end
	return str
end

fboSize = 256
fbo = FBO.new{width=fboSize; height=fboSize; useDepth=true;}

objs = table.new()

Object = {}
function Object:new(color, render, update)
	if not update then update = function() end end
	local obj = {
		pos={0,0,0};
		color=color;
		render=render;
		update=update;
	}


	obj.cubeMapTex = TexCube.new{
		width=fboSize;
		height=fboSize;
		wrap={s=gl.GL_CLAMP, t=gl.GL_CLAMP};
		generateMipmap=true;
		-- for uploads does the data and type have to remain the same?
		format = gl.GL_RGB;
		type = gl.GL_UNSIGNED_BYTE;
		minFilter=gl.GL_LINEAR;
		magFilter=gl.GL_LINEAR;
	}

	objs:insert(obj)
	return obj
end

-- draw functions
function drawTeapot() glush.drawOBJ('cubefbo/teapot.obj') end
-- update functions
function updateRotateByTime(m)
	return function(self)
		-- update code.  once per cycle
		local theta = glush.timer()*.5
		local rad = 10
		local x = rad*math.cos(theta)
		local y = rad*math.sin(theta)
		for i=1,3 do
			self.pos[i] = m[i][1] * x + m[i][2] * y
		end
	end
end

Object:new({.5,.5,.5}, glush.drawSphere)
Object:new({1,0,0}, drawTeapot, updateRotateByTime{{1,0},{0,0},{0,1}} )
Object:new({0,1,0}, drawTeapot, updateRotateByTime{{0,-1},{0,0},{1,0}} )
--[[
Object:new({0,0,1}, drawTeapot, updateRotateByTime{{-1,0},{0,0},{0,-1}} )
Object:new({1,1,0}, drawTeapot, updateRotateByTime{{0,1},{0,0},{-1,0}} )
Object:new({1,0,1}, drawTeapot, updateRotateByTime{{1,0},{0,1},{0,0}} )
Object:new({0,1,1}, drawTeapot, updateRotateByTime{{-1,0},{0,-1},{0,0}} )
--]]


--[[
cubemap fbo ...
how this will work

1. render lots of objects around the scene
2. for one of them create our empty cubemap ... 
--]]


-- in transition:

-- i set up enable/disable to accept as many params as you can stuff in it
-- i know, i'm stretching my jurisdiction.
gl.glEnable(gl.GL_DEPTH_TEST, gl.GL_NORMALIZE, gl.GL_CULL_FACE)

gl.glHint(gl.GL_GENERATE_MIPMAP_HINT, gl.GL_NICEST)	-- GL_NICEST or GL_FASTEST

gl.glLightModel{
	localViewer = true;
	ambient = {0,0,0,1};
	colorControl = gl.GL_SEPARATE_SPECULAR_COLOR;
}

gl.glLight{
	[0] = {		-- im debating better ways to reference which light # ... esp with lua 1-indexes
		ambient = {1,1,1,1};
		diffuse = {1,1,1,1};
		specular = {1,1,1,1};
		position = {0,0,-1,1};
		enable = true;
	}
}

gl.glMaterial{
	front = {
		emission = {0,0,0,0};
		ambient = {0,0,0,1};
		diffuse = {1,1,1,1};
		specular = {1,1,1,1};
		shininess = 10;
	}
}

gl.glColorMaterial(gl.GL_FRONT_AND_BACK, gl.GL_DIFFUSE)
gl.glEnable(gl.GL_LIGHTING, gl.GL_COLOR_MATERIAL)

cubeTexShaderCode = io.readfile('cubefbo/cubemap.shader')
cubeTexShader = ShaderProgram.new{
	vertexCode = defines('VERTEX_SHADER')..cubeTexShaderCode;
	fragmentCode = defines('FRAGMENT_SHADER')..cubeTexShaderCode;
	uniforms = { cubemap = 0;};
}

-- callback from the main cheese
function glush.display()
	-- update objects
	update()
	
	-- render code.
	-- update cubemaps
	updateCubemaps()
	-- then draw the scene
	render()
end

function update() 	
	for i,obj in ipairs(objs) do
		obj:update()
	end
end

function updateCubemaps() 
	-- do i set up the cubemap and then bind the framebuffer to its appropriate side?
	-- just like with sw mode ...
	gl.glPushAttrib(gl.GL_VIEWPORT_BIT)
	gl.glViewport(0,0,fboSize,fboSize)

	--save matrices
	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glPushMatrix()
		-- set up the new projection while you're here
	gl.glLoadIdentity()
	gl.glFrustum(-.1,.1,-.1,.1,.1,100)
	
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glPushMatrix()
	gl.glLoadIdentity()

	for i,obj in ipairs(objs) do
		updateCubemap(obj)
	end
	
	--restore matrices
	gl.glMatrixMode(gl.GL_PROJECTION)
	gl.glPopMatrix()
	gl.glMatrixMode(gl.GL_MODELVIEW)
	gl.glPopMatrix()

	gl.glPopAttrib()
end

function updateCubemap(obj)
	fbo:bind()
	for i=0,5 do
		fbo:setColorAttachmentTex2D(0, obj.cubeMapTex.id, gl.GL_TEXTURE_CUBE_MAP_POSITIVE_X+i)
		fbo:bind()
		assert(fbo:check())

		gl.glClear(gl.GL_COLOR_BUFFER_BIT + gl.GL_DEPTH_BUFFER_BIT)
		gl.glLoadIdentity()
		if i==0 then
			gl.glRotate(180,0,0,1)	-- 180 twist
			gl.glRotate(90,0,1,0) -- turn to the right
		elseif i==1 then
			gl.glRotate(180,0,0,1)	-- 180 twist
			gl.glRotate(90,0,-1,0)	-- left
		elseif i==2 then
			gl.glRotate(90,-1,0,0)	-- up
		elseif i==3 then
			gl.glRotate(90,1,0,0)	-- down
		elseif i==4 then
			gl.glRotate(180,0,0,1)	-- 180 twist
			gl.glRotate(180,0,1,0)	-- back
		elseif i==5 then
			gl.glRotate(180,0,0,1)	-- 180 twist
			--forward
		end
		gl.glTranslate(-obj.pos[1], -obj.pos[2], -obj.pos[3])
		render(obj)
		fbo:unbind()
	end
end

function render(ignore)
	for i,obj in ipairs(objs) do
		if obj ~= ignore then
			gl.glPushMatrix()
			gl.glTranslate(obj.pos)
			
			gl.glEnable(gl.GL_TEXTURE_CUBE_MAP)
			gl.glBindTexture(gl.GL_TEXTURE_CUBE_MAP, obj.cubeMapTex.id)
			cubeTexShader:use()
			
			gl.glVertexAttrib('reflectivity', 1.)
			gl.glColor(obj.color)
			obj:render()
			
			ShaderProgram.useNone()
			gl.glDisable(gl.GL_TEXTURE_CUBE_MAP)

			gl.glPopMatrix()
		end
	end
end
